/**
 * gen_hypohamiltonian.c
 *
 * GenHypohamiltonian is a generator for hypohamiltonian and almost hypohamiltonian graphs.
 *
 * The latest version of gen_hypohamiltonian can be found here:
 * http://caagt.ugent.be/hypoham/
 * 
 * Author: Jan Goedgebeur (jan.goedgebeur@ugent.be)
 * In collaboration with Carol Zamfirescu
 *
 * -----------------------------------
 * 
 * Copyright (c) 2016-2018 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Compile with:
 *
 * cc -DWORDSIZE=32 -DMAXN=WORDSIZE -mpopcnt -O4 gen_hypohamiltonian.c planarity.c nautyW1.a -o gen_hypohamiltonian
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
//#include "nauty.h"
#include "planarity.h" /* which includes nauty.h */

#define VERSION "1.2.1 - September 18 2018"


/*************************Defines and global variables*************************/

/******************************Configuration***********************************/

//For option mod: splitting when <x> edges are added
//3 seems to be a good compromise for all cases
#define DESIRED_SPLITLEVEL 3

/***************You should not change anything below this line*****************/

//For the timing
#define time_factor sysconf(_SC_CLK_TCK) 

/*****************************Some useful macros*******************************/

#if WORDSIZE == 64
#define BIT(i) (1ULL << (i))
#else //i.e. WORDSIZE == 32
#define BIT(i) (1U << (i))
#endif

#define MAX(a, b) (((a) >= (b)) ? (a) : (b))
#define MIN(a, b) (((a) <= (b)) ? (a) : (b))

//size of multicode is nv + ne        
#define MAX_MUTLICODE_SIZE(a) ((a) + (a)*((a)-1)/2)


//Uncomment to use builtin popcount (else uses nauty popcount)
//Need to compile with gcc -mpopcnt or -msse4.2 to enable __builtin_popcount!
//USE_BUILTIN_POPCOUNT is significantly faster for 64bit
#define USE_BUILTIN_POPCOUNT

#ifdef USE_BUILTIN_POPCOUNT

#if WORDSIZE == 64
#define POPC(x) (__builtin_popcountll(x)) 
#else //i.e. WORDSIZE == 32
#define POPC(x) (__builtin_popcount(x)) 
#endif

#define POPCLL(x) (__builtin_popcountll(x)) 

#else
#define POPC(x) POPCOUNT(x) /* needs nauty.h */
#endif

/***********************Methods for splay-tree*****************************/

//Methods for the splay-tree (see splay.c for more information). 

typedef struct sp {
    //graph canon_form[MAXN*MAXM];
    graph* canon_form; //Is a pointer, should use malloc for this; isn't really faster but uses fewer memory!
    struct sp *left, *right, *parent;
} SPLAYNODE;


//SPLAYNODE *worklist = NULL;
//list_of_worklists[i] is the worklist for graphs with i vertices
SPLAYNODE *list_of_worklists[MAXN+1] = {NULL};

/* extra arguments to pass to scan procedure */
#define SCAN_ARGS

/* what scan procedure should do for each node p */
#define ACTION(p) free_canon_form_node(p)

/* extra arguments for the insertion procedure */
#define INSERT_ARGS , graph *canong, int *is_new_node

/* how to compare the key of INSERT_ARGS to the key of node p */
#define COMPARE(p) comparenodes(canong, p)

/* what to do for a new node p */
#define NOT_PRESENT(p) new_splaynode(p, canong, is_new_node)

/* what to do for an old code p */
#define PRESENT(p) old_splaynode(p, is_new_node)


void new_splaynode();
void old_splaynode();

#include "splay.c"

/******************************************************************************/

//+1 needed for minibaum representation
typedef unsigned char GRAPH[MAXN+1][MAXN];
typedef unsigned char ADJACENCY[MAXN+1];
typedef unsigned char EDGE[2];

//The graph which is currently under construction
GRAPH current_graph;
ADJACENCY adj;

int nv = 0; //Nv of current_graph
int maxnv = 0;

//For nauty
#define WORKSIZE (50 * MAXM)

int lab[MAXN], ptn[MAXN], orbits[MAXN];
static DEFAULTOPTIONS_GRAPH(options);
statsblk stats;
setword workspace[WORKSIZE];

graph nautyg[MAXN*MAXM];


#define MAXE_HYPO (MAXN*(MAXN-1) / 2)

//List of all edges for which the result is still nonham and which doesn't increase the max deg
EDGE candidate_edges[MAXE_HYPO];
int num_candidate_edges = 0;

int num_edges_with_given_endpoint[MAXN];

//For hamiltonicity test

static int current_path[MAXN];
static int current_pathsize = 0;
static int cycle_found = 0;

static int component_nr_of_vertex[MAXN];
int current_type_A_component;


int type_B_obstruction_found = 0;
int degree_in_graph_induced_by_partition[MAXN];

setword type_B_W_vertices = 0;


int type_C_obstruction_found = 0;


int test_if_trianglefree = 0;

int test_if_planar = 0;


int test_girth = 0;
int girth = 0;

int impose_max_deg = 0;
int max_deg;

int impose_min_max_deg = 0;
int min_max_deg;

//Special hack doesn't allow vertices with degree larger than hub
int test_almost_hypoham_special = 0;

int test_almost_hypoham_complete = 0;

int num_edges_added = 0;

int max_num_edges_added = 0;


int test_platypus = 0;


//For induced cycle test
static unsigned long long int possible_cycle_vertices = 0;
static int contains_forbidden_cycle = 0;

//static int current_path[MAXN];
//static int current_pathsize = 0;
static setword current_path_bitvector;

/******************************************************************************/

/* Some marks */
#include <limits.h> //Necessary for marks
unsigned int marks[MAXN];

#define MAXVAL INT_MAX - 1
static int markvalue = MAXVAL;
#define RESETMARKS {int mki; if ((markvalue += 1) > MAXVAL) \
      { markvalue = 1; for (mki=0;mki<MAXN;++mki) marks[mki]=0;}}
#define MARK(v) marks[v] = markvalue
#define UNMARK(v) marks[v] = markvalue - 1
#define ISMARKED(v) (marks[v] == markvalue)


/**************** Start variables for hypoham test ****************************/

//This hypoham test was implemented and kindly provided to us by Gunnar Brinkmann

//Minibaum graph (here g[0][0]=nv)
GRAPH current_graph_mb;
ADJACENCY adj_mb;

#define leer 255 


int kantenzahl, maxvalence, welchergraph, codelaenge, schreiben=0, alle=0;
int filter=0, filter_h=0, filter_nh=0;
unsigned char knotenzahl, startvertex_HC;

GRAPH forced_edges;
ADJACENCY anzahl_forced;
unsigned long int original[MAXN+1]; // the graph as bitvector

#define bit(i) ((1UL)<<(i))
#define DELBIT(a,i) ((a)&=(~bit(i)))
#define SETBIT(a,i) ((a)|=(bit(i)))

unsigned long int _gbu;
#define FORALLELEMENTS(x,y) for (_gbu=x, y= __builtin_ffsl(_gbu); y; _gbu&=(_gbu-1), y= __builtin_ffsl(_gbu))


#define LESS_THAN_2BIT(a) (((a)&((a)-1))==0UL)

/**************** End Variables for hypoham test ******************************/


/* Specific parameters for modulo */
static int modulo = 1;
static int rest = 0;
static int splitcounter = 0;
static int splitlevel = 0; /* For modolo, determines at what level/depth of the recursion tree the calculation should be split */

/******************************************************************************/

//For statistics
static unsigned long long int num_graphs_generated_iso = 0;
static unsigned long long int num_graphs_generated_noniso = 0;
static unsigned long long int num_graphs_written = 0;


static unsigned long long int times_contains_type_A_obstruction = 0;
static unsigned long long int times_doesnt_contain_type_A_obstruction = 0;

static unsigned long long int times_contains_deg_two_vertex = 0;
static unsigned long long int times_doesnt_contain_deg_two_vertex = 0;

static unsigned long long int times_contains_deg_three_triangle_vertex = 0;
static unsigned long long int times_doesnt_contain_deg_three_triangle_vertex = 0;

static unsigned long long int times_contains_type_B_obstruction = 0;
static unsigned long long int times_doesnt_contain_type_B_obstruction = 0;

static unsigned long long int times_contains_type_C_obstruction = 0;
static unsigned long long int times_doesnt_contain_type_C_obstruction = 0;

static unsigned long long int times_no_lemmas_applicable = 0;

static unsigned long long int times_except_vertex_ham = 0;
static unsigned long long int times_except_vertex_not_ham = 0;

static unsigned long long int times_hamiltonian = 0;
static unsigned long long int times_not_hamiltonian = 0;

static unsigned long long int times_planar = 0;
static unsigned long long int times_not_planar = 0;


static int number_of_indep_sets_found[MAXN+1];

/*********************************Methods**************************************/

static void extend(int max_deg, int num_w_vertices, setword w_bitvector);

/******************************************************************************/

/**
 * Prints the given nauty graph.
 */
static void
printgraph_nauty(graph *g, int current_number_of_vertices) {
    int i;
    fprintf(stderr, "Printing graph nauty:\n");
    for(i = 0; i < current_number_of_vertices; i++) {
        fprintf(stderr, "%d :", i);
        set *gv = GRAPHROW1(g, i, MAXM);
        int neighbour = -1;
        while((neighbour = nextelement(gv, MAXM, neighbour)) >= 0) {
            fprintf(stderr, " %d", neighbour);
        }
        fprintf(stderr, "\n");
    }
}

/******************************************************************************/

static void
printgraph_minibaum(GRAPH g, ADJACENCY adj, int nv) {
    int i, j;
    fprintf(stderr, "Printing graph: \n");
    for(i = 1; i <= nv; i++) {
        fprintf(stderr, "%d :", i);
        for(j = 0; j < adj[i]; j++)
            fprintf(stderr, " %d", g[i][j]);
        fprintf(stderr, "\n");
    }
}

/******************************************************************************/

static void
printgraph(GRAPH g, ADJACENCY adj, int nv) {
    int i, j;
    fprintf(stderr, "Printing graph: \n");
    for(i = 0; i < nv; i++) {
        fprintf(stderr, "%d :", i);
        for(j = 0; j < adj[i]; j++)
            fprintf(stderr, " %d", g[i][j]);
        fprintf(stderr, "\n");
    }
}

/*********************Methods for splay tree*********************************/

void new_splaynode(SPLAYNODE *el, graph *canong, int *is_new_node) {
    el->canon_form = canong;
    //el->av_colours_canon = av_cols;
    //memcpy(el->canon_form, canong, sizeof(graph) * nv);
    *is_new_node = 1;
}

/****************************************/

void old_splaynode(SPLAYNODE *el, int *is_new_node) {
    *is_new_node = 0;
    //fprintf(stderr, "Graph was already present. Nv: %d\n", nv);
}

/****************************************/

int comparenodes(graph *canong, SPLAYNODE *list) {
    return memcmp(canong, list->canon_form, nv * sizeof (graph));
}

/****************************************/

free_canon_form_node(SPLAYNODE *el) {
    free(el->canon_form);
}

/******************************************************************************/

/* Assumes that MAXM is 1 (i.e. one_word_sets) */
//This is slightly faster...
int
nextelement1(set *set1, int m, int pos) {
    setword setwd;

    if(pos < 0) setwd = set1[0];
    else setwd = set1[0] & BITMASK(pos);

    if(setwd == 0) return -1;
    else return FIRSTBIT(setwd);
}

/******************************************************************************/

/*
 * Converts g to a graph in the nauty format and stores the resulting graph
 * in nauty_graph.
 */
static void
copy_nauty_graph(GRAPH g, ADJACENCY adj, int num_vertices, graph *nauty_graph) {
    int i, j;
    for(i = 0; i < num_vertices; i++) {
        nauty_graph[i] = 0;
        for(j = 0; j < adj[i]; j++)
            ADDELEMENT1(&nauty_graph[i], g[i][j]);
    }

}

/******************************************************************************/

/**
 * Code the graph using the multicode format and returns the length of the code.
 */
int code_multicode(unsigned char code[], GRAPH g, ADJACENCY adj, int num_of_vertices) {
    int i, j;
    int codelength = 0;
    code[codelength++] = num_of_vertices;
    for(i = 0; i < num_of_vertices - 1; i++) {
        for(j = 0; j < adj[i]; j++)
            if(g[i][j] > i)
                code[codelength++] = g[i][j] + 1;
        code[codelength++] = 0;
    }

    return codelength;
}

/******************************************************************************/

//Outputs the graph to stdout in multicode
static void
output_graph(GRAPH g, ADJACENCY adj, int num_vertices) {
    
    unsigned char codebuffer[MAX_MUTLICODE_SIZE(num_vertices)];
    int codelength = code_multicode(codebuffer, g, adj, num_vertices);

    if(fwrite(codebuffer, sizeof (unsigned char), codelength, stdout)
            < sizeof (unsigned char) * codelength) {
        fprintf(stderr, "Error: couldn't write graph! \n\n");
        exit(1);
    }

    num_graphs_written++;
}

/******************************************************************************/

/**
 * Adds a new vertex and connects it will all vertices of
 * the vertex set. 
 */
static void
expand_vertex_set_no_remove(unsigned char dominating_set[], int set_size) {
    nautyg[nv] = 0;
    int i;
    for(i = 0; i < set_size; i++) {
        current_graph[dominating_set[i]][adj[dominating_set[i]]++] = nv;
        current_graph[nv][i] = dominating_set[i];
        ADDELEMENT1(&nautyg[dominating_set[i]], nv);
        ADDELEMENT1(&nautyg[nv], dominating_set[i]);
    }
    adj[nv] = set_size;

    nv++;
}

/******************************************************************************/

/**
 * Reduction of expand_dominating_set_no_remove()
 */
static void
reduce_vertex_set_no_remove(unsigned char dominating_set[], int set_size) {
    nv--;

    int i;
    for(i = 0; i < set_size; i++) {
        adj[dominating_set[i]]--;
        DELELEMENT1(&nautyg[dominating_set[i]], nv);
    }

}

/******************************************************************************/

//Important: it is assumed that the current graph has girth at least "girth" = forbiddencyclesize
static void
search_cycle(int forbiddencyclesize) {
    
    int previous_vertex = current_path[current_pathsize-1];
    int i;
    for(i = 0; i < adj[previous_vertex]; i++) {
        int next_vertex = current_graph[previous_vertex][i];
        if(!ISMARKED(next_vertex)) {
            if(current_pathsize < forbiddencyclesize - 1) {
                MARK(next_vertex);
                current_path[current_pathsize++] = next_vertex;

                //recursion
                search_cycle(forbiddencyclesize);
                if(cycle_found)
                    return;                

                UNMARK(next_vertex);
                current_pathsize--;
            } else if(current_pathsize == forbiddencyclesize - 1 &&
                    ISELEMENT1(&nautyg[next_vertex], current_path[0])) {

                current_path[current_pathsize++] = next_vertex;
                
                cycle_found = 1;

                return;
            }
        }
    }
}

/******************************************************************************/

//The algorithm by Gunnar Brinkmann is significantly faster than this brute-force algorithm
int contains_hamiltonian_cycle_old() {
    //Start from vertex with smallest degree
    int min_deg = MAXN;
    int startvertex;
    int i;
    for(i = 0; i < nv; i++)
        if(adj[i] < min_deg) { //This seems to be the best choice for a startvertex!
        //if(adj[i] <= min_deg) {
            startvertex = i;
            min_deg = adj[i];
        }
    
    if(min_deg == MAXN) {
        fprintf(stderr, "Error: invalid min deg!\n");
        exit(1);
    }
    
    if(min_deg < 2) {
        fprintf(stderr, "Error: min deg is too small!\n");
        exit(1);
    }
    
    
/*
    RESETMARKS;

    current_path[0] = startvertex;
    current_pathsize = 1;
    MARK(current_path[0]);

    cycle_found = 0;

    //recursion
    search_cycle(nv);    

    return cycle_found;
*/
    
    //This helps significantly!

    RESETMARKS;
    cycle_found = 0;
    MARK(startvertex);
    
    //Min deg is at least 2!
    int j, k;
    for(j = 0; j < adj[startvertex]; j++)
        for(k = j + 1; k < adj[startvertex]; k++) {
            int nbr1 = current_graph[startvertex][j];
            int nbr2 = current_graph[startvertex][k];
            current_path[0] = nbr1;
            current_path[1] = startvertex;
            current_path[2] = nbr2;
            MARK(current_path[0]);
            //MARK(current_path[1]);
            MARK(current_path[2]);
            current_pathsize = 3;

            search_cycle(nv);

            if(cycle_found)
                return 1;
            
            UNMARK(current_path[0]);
            //UNMARK(current_path[1]);
            UNMARK(current_path[2]);
            
        }    
    
    return 0;
    
}


/*************************TESTE_FORCED*******************/

int teste_forced(GRAPH graph,ADJACENCY adj)

{
int i,j,k;

for (i=1; i<= graph[0][0]; i++)
  for (j=0; j<anzahl_forced[i]; j++)
    { for (k=0; (k<adj[i]) && (graph[i][k] != forced_edges[i][j]); k++);
      if (k==adj[i]) { fprintf(stderr,"Kante %d-%d nicht enthalten.\n",i,forced_edges[i][j]);
		       return(0); }
    }
return(1);
}

/*******************WEITERHAMMI**************************/

void weiterhammi(GRAPH graph,ADJACENCY adj, int wo, char marks[],
		 int tiefe, int *gefunden,unsigned char zykel[],unsigned long int left_to_visit)

{
  int komplett=0, j, i;
int next;
 unsigned long int nbset;

 //for (i=0;i<tiefe;i++) fprintf(stderr,"%d-",zykel[i]); fprintf(stderr,"\n");

if (tiefe==graph[0][0]-1)
  {
    if (!alle || anzahl_forced[0] || (wo > zykel[1]))
  /* da jeder Zykel nur einmal gefunden werden soll, wird im Falle, dass alle Zykel
     konstruiert werden sollen nur ueberprueft, falls die erste Kante gezwungen war, 
     oder zumindest lexikographisch kleiner als die letzte. */
      { if (anzahl_forced[wo]==2)
	  { if ((forced_edges[wo][0]==zykel[0]) || (forced_edges[wo][1]==zykel[0]))
	      { komplett=1; zykel[tiefe]=wo; }
	  }
      else /* wenn (anzahl_forced[wo]==1), so geht die gezwungene Kante entweder zu
	      zykel[0] (dann ist es ja OK), oder zu einem anderen Knoten, von dem
	      man dann zwangsweise gerade gekommen ist -- es muss also nur ueberprueft werden,
	      ob eine Kante wo->zykel[0] existiert */
	for (j=0; j<adj[wo]; j++) if (graph[wo][j]==zykel[0]) { komplett=1; 
								j=adj[wo];
								zykel[tiefe]=wo; }
	if (komplett)
	  { (*gefunden)++;
	    if (schreiben)
	      { fprintf(stderr,"Hamiltonzykel: "); 
		for (j=0; j<=tiefe; j++) fprintf(stderr,"%d->",zykel[j]);
		fprintf(stderr,"\n"); }
	  }
      }
    return;
  }

left_to_visit &= ~bit(wo);

  if ((original[startvertex_HC]&left_to_visit)==0UL) return;
  if ((original[wo]&left_to_visit)==0UL) return;

  //writeset(left_to_visit);
  //fprintf(stderr,"wordsize=%d\n",WORDSIZE);

  nbset=left_to_visit | bit(startvertex_HC) | bit(wo);
  FORALLELEMENTS(left_to_visit,i) 
    { if (LESS_THAN_2BIT(original[i-1]&nbset)) return; }

if ((anzahl_forced[wo]==2) || ((anzahl_forced[wo]==1) && (zykel[tiefe-1] != forced_edges[wo][0])))
  /* Es gibt eine gezwungene Kante */
  { if (zykel[tiefe-1] != forced_edges[wo][0]) next=forced_edges[wo][0];
      else next=forced_edges[wo][1];
    if (!marks[next])
      {
	marks[wo]=1;
	zykel[tiefe]=wo;
	weiterhammi(graph,adj,next,marks,tiefe+1,gefunden,zykel,left_to_visit);
	marks[wo]=0;
      }
    return;
  }

/* sonst gibt es keine gezwungene Kante -- es muss sogar eine ungezwungene Kante
   genommen werden: */

marks[wo]=1;
zykel[tiefe]=wo;
for (j=0; (j<adj[wo]) && (!(*gefunden) || alle) ; j++)
  { next=graph[wo][j];
    if ( !marks[next] && (anzahl_forced[next]<2))
      weiterhammi(graph,adj,next,marks,tiefe+1,gefunden,zykel,left_to_visit);
  }
marks[wo]=0;
return;
    }


/*************************HAMMI_START*******************/

//Returns 1 if the graph is hamiltonian, else returns 0
//This hypoham test was implemented and kindly provided to us by Gunnar Brinkmann

//This test is significantly faster than the brute-force algorithm contains_hamiltonian_cycle_old()

int hammi_start(GRAPH graph,ADJACENCY adj)

{

int start, next, j, max, best, i;
char marks[MAXN+1];
unsigned char zykel[MAXN];
int gefunden;

unsigned long int left_to_visit;

 left_to_visit= ((1UL<<(graph[0][0]))-1)<<1;

 for (i=1;i<=graph[0][0];i++)
   { original[i]=0UL;
     for (j=0;j<adj[i]; j++) original[i] |= bit(graph[i][j]);
   }

//if (schreiben) schreibegraph(graph,adj);

if (anzahl_forced[0]) { while (anzahl_forced[start]==0) start++;
			if (teste_forced(graph,adj)==0) return(0); }
 else
   {
     best=adj[1], start=1;
     for (i=2; i<=graph[0][0]; i++)
       if (adj[i]<best)
	 { start=i; best=adj[start]; }
   }

for (j=0; j<=MAXN; j++) marks[j]=0;

left_to_visit&= ~bit(start);
startvertex_HC=start;
marks[start]=1;
zykel[0]=start;
gefunden=0;
if (anzahl_forced[0]) 
  { next=forced_edges[start][0];
    weiterhammi(graph,adj,next,marks,1,&gefunden,zykel,left_to_visit);
    return(gefunden);
  }

else
  { for (j=max=0; j<adj[start] ; j++) if (graph[start][j]>max) max=graph[start][j];
    for (j=0; j<adj[start] && (!gefunden || alle); j++) 
      /*  da ja ZWEI Kanten hier starten muessen, kann
	  obda zuerst die lexikographisch kleinere kommen */
      { //fprintf(stderr,"--- %d %d  gefunden %d alle %d max %d \n",start,graph[start][j],gefunden,alle,max);
	if ((next=graph[start][j]) != max)
	  weiterhammi(graph,adj,next,marks,1,&gefunden,zykel,left_to_visit); 
      }
	return(gefunden);
  }
}

/******************************************************************************/

int contains_hamiltonian_cycle() {
    //Is (slightly) faster when just making current_graph again than doing it once
    //and adapting determine_longest_induced_path_recursion() to use the n+1 repres
    current_graph_mb[0][0] = nv;
    int i, j;
    for(i = 1; i <= nv; i++) {
        adj_mb[i] = adj[i-1];
        for(j = 0; j < adj[i-1]; j++)
            current_graph_mb[i][j] = current_graph[i-1][j] + 1;

        //Is necessary as it is used in the colouring routines
        current_graph_mb[i][j] = leer;
    }    
    
    return hammi_start(current_graph_mb, adj_mb);
}

/******************************************************************************/

/**
 * G is hypohamiltonian if G is non-hamiltonian but
 * G - v (for all V \in V(G)) contains a hamilton cycle.
 */
//Important: it is assumed that G is nonhamiltonian!
int is_hypohamiltonian() {

/*
    //First verify that inputgraph is non-hamiltonian
    if(contains_hamilton_cycle()) {
        fprintf(stderr, "Error: inputgraphs are assumed to be non-hamiltonian!\n");
        exit(1);
    }
*/
    
    //TODO: could use a more efficient ham algorithm for this, but this is not a bottleneck!

    int i;
    for(i = 0; i < nv; i++)
        if(adj[i] < 3)
            return 0; //hypoham graphs are 3-connected so have mindeg at least 3

    for(i = 0; i < nv; i++) {
        RESETMARKS;
        MARK(i); //i is forbidden

        //Best to start with vertex of min deg!
        //But won't make much difference since this is not a bottleneck!
        int startvertex = (i + 1) % nv;

        current_path[0] = startvertex;
        current_pathsize = 1;
        MARK(startvertex);

        cycle_found = 0;
        search_cycle(nv - 1);

        if(!cycle_found)
            return 0; //Not hypohamiltonian
    }

    return 1;

}

/******************************************************************************/

static int path_found = 0;

//Important: it is assumed that the current graph has girth at least "girth" = forbiddencyclesize
static void
search_path(int first_vertex, int previous_vertex, int current_pathsize, int forbiddencyclesize) {
    
    //int previous_vertex = current_path[current_pathsize-1];
    int i;
    for(i = 0; i < adj[previous_vertex]; i++) {
        int next_vertex = current_graph[previous_vertex][i];
        if(!ISMARKED(next_vertex)) {
            if(current_pathsize < forbiddencyclesize - 1) {
                MARK(next_vertex);
                //current_path[current_pathsize++] = next_vertex;

                //recursion
                search_path(first_vertex, next_vertex, current_pathsize + 1, forbiddencyclesize);
                if(path_found)
                    return;                

                UNMARK(next_vertex);
                //current_pathsize--;
            } else if(current_pathsize == forbiddencyclesize - 1) {
                    //&& is_neighbour(next_vertex, first_vertex)) {

                //current_path[current_pathsize++] = next_vertex;
                
                path_found = 1;

                return;
            }
        }
    }
}

/******************************************************************************/

/**
 * G is hypohamiltonian if G is non-hamiltonian but
 * G - v (for all V \in V(G)) contains a hamilton cycle.
 */
int is_hypotraceable() {
    
/*
    //First verify that inputgraph is non-traceable
    if(contains_hamilton_path()) {
        fprintf(stderr, "Error: inputgraphs are assumed to be non-hamiltonian!\n");
        exit(1);
    }
*/
    
    //TODO: could use a more efficient ham algorithm for this, but this is not a bottleneck!
    
    
    int i;
    for(i = 0; i < nv; i++) {
        //RESETMARKS;
        //MARK(i); //i is forbidden
        
        path_found = 0;        
        
        int j;
        for(j = 0; j < nv && !path_found; j++)
            if(j != i) {
                RESETMARKS;

                MARK(i); //i is forbidden
                MARK(j);
                
                path_found = 0;

                //recursion
                search_path(j, j, 1, nv - 1);

            }
        if(!path_found)
            return 0; //i.e. G-i is nontraceable, so it's not hypotraceable!
    }
    
    return 1;

}

/******************************************************************************/

/**
 * G is hypohamiltonian if G is non-hamiltonian but
 * G - v (for all V \in V(G)) contains a hamilton cycle.
 */
//Important: it is assumed that G is nonhamiltonian!
int is_almost_hypohamiltonian() {

/*
    //First verify that inputgraph is non-hamiltonian
    if(contains_hamilton_cycle()) {
        fprintf(stderr, "Error: inputgraphs are assumed to be non-hamiltonian!\n");
        exit(1);
    }
*/

    int i;
    for(i = 0; i < nv; i++)
        if(adj[i] < 3)
            return 0; //almost hypoham graphs are 3-connected so have mindeg at least 3
    
    int num_exceptional_vertices = 0;

    for(i = 0; i < nv; i++) {
        RESETMARKS;
        MARK(i); //i is forbidden

        //Best to start with vertex of min deg!
        //But won't make much difference since this is not a bottleneck!
        int startvertex = (i + 1) % nv;

        current_path[0] = startvertex;
        current_pathsize = 1;
        MARK(startvertex);

        cycle_found = 0;
        search_cycle(nv - 1);

        if(!cycle_found) {
            num_exceptional_vertices++;
            if(num_exceptional_vertices > 1)
                return 0;
        }
    }

    return num_exceptional_vertices == 1;

}

/******************************************************************************/

static void
add_edge(unsigned char v, unsigned char w) {
    current_graph[v][adj[v]] = w;
    current_graph[w][adj[w]] = v;
    adj[v]++;
    adj[w]++;
    ADDELEMENT1(&nautyg[v], w);
    ADDELEMENT1(&nautyg[w], v);
    num_edges_added++;
}

/******************************************************************************/

//Important: it is assumed that v and w are at the end of the adjacency list!
static void
remove_edge(unsigned char v, unsigned char w) {
    adj[v]--;
    adj[w]--;
    DELELEMENT1(&nautyg[v], w);
    DELELEMENT1(&nautyg[w], v);  
    num_edges_added--;
}

/******************************************************************************/

//It is assumed that current_pathlength > 0
static void
contains_forbidden_induced_cycle_recursion(int forbiddencyclesize) {
    if(current_pathsize >= forbiddencyclesize)
        return;
    
    int previous_vertex = current_path[current_pathsize-1];
    int i;
    for(i = 0; i < adj[previous_vertex]; i++) {
        int next_vertex = current_graph[previous_vertex][i];

        if(ISELEMENT1(&possible_cycle_vertices, next_vertex)) { //marks seem to be faster than bv here
        //if(ISELEMENT1(&possible_cycle_vertices, next_vertex) && !ISMARKED(next_vertex)) { //marks seem to be faster than bv here
        //if((BIT(next_vertex) & current_path_bitvector) == 0) {
            if(current_pathsize < forbiddencyclesize - 1 
                    //Doesn't have to be an induced cycle, but it is best to be as restrictive as possible as usually it does not contain a C5
                    && (*GRAPHROW1(nautyg, next_vertex, MAXM) & current_path_bitvector) == BITT[previous_vertex]) {
                    //(neighbourhood[next_vertex] & current_path_bitvector) == BIT(previous_vertex)) {
                //MARK(next_vertex);
                DELELEMENT1(&possible_cycle_vertices, next_vertex);
                current_path[current_pathsize++] = next_vertex;
                //current_path_bitvector |= BIT(next_vertex);
                ADDELEMENT1(&current_path_bitvector, next_vertex);

                //recursion
                contains_forbidden_induced_cycle_recursion(forbiddencyclesize);
                if(contains_forbidden_cycle)
                    return;                

                //UNMARK(next_vertex);
                ADDELEMENT1(&possible_cycle_vertices, next_vertex);
                current_pathsize--;
                //current_path_bitvector &= ~BIT(next_vertex);
                DELELEMENT1(&current_path_bitvector, next_vertex);
            } else if(current_pathsize == forbiddencyclesize - 1 &&
                    //(neighbourhood[next_vertex] & current_path_bitvector) 
                    //== (BIT(previous_vertex) | BIT(current_path[0]))) {
                    //Doesn't have to be an induced cycle, but it is best to be as restrictive as possible as usually it does not contain a C5
                    //(*GRAPHROW1(nautyg, next_vertex, MAXM) & BITT[current_path[0]]) != 0) { //Moet zelfs niet induced zijn!
                    (*GRAPHROW1(nautyg, next_vertex, MAXM) & current_path_bitvector) 
                    == (BITT[previous_vertex] | BITT[current_path[0]])) {
                //Induced cycle formed of size current_pathsize + 1
                //fprintf(stderr, "cycle found of size %d\n", current_pathsize + 1);
                if(current_pathsize + 1 == forbiddencyclesize) {
                    contains_forbidden_cycle = 1;
                    return;
                }
                
            }
        }
    }
}

/******************************************************************************/

/**
 * Returns 1 if the graph contains an induced cycle of size forbiddencyclesize,
 * else returns 0.
 */
static int
contains_forbidden_induced_cycle(int forbiddencyclesize) {
    contains_forbidden_cycle = 0;
    
    possible_cycle_vertices = ALLMASK(nv);
    
    int i;
    for(i = 0; i < nv; i++) {
        current_path[0] = i;
        current_pathsize = 1;
        current_path_bitvector = 0;
        ADDELEMENT1(&current_path_bitvector, i);
        DELELEMENT1(&possible_cycle_vertices, i);        
        
        contains_forbidden_induced_cycle_recursion(forbiddencyclesize);
        
        if(contains_forbidden_cycle)
            return 1;        
        
        //No need to unmark!
        
    }

    return 0;
}


/******************************************************************************/

/**
 * Returns 1 if the graph contains an induced cycle of size forbiddencyclesize,
 * else returns 0.
 */
//Since the parents were C-free, the last edge must be in the forbidden cycle
static int
contains_forbidden_induced_cycle_last_edge(unsigned char v1, unsigned char v2, int forbiddencyclesize) {
    contains_forbidden_cycle = 0;

    possible_cycle_vertices = ALLMASK(nv);

    current_path[0] = v1;
    current_path[1] = v2;
    current_pathsize = 2;
    current_path_bitvector = 0;
    ADDELEMENT1(&current_path_bitvector, v1);
    DELELEMENT1(&possible_cycle_vertices, v1);
    ADDELEMENT1(&current_path_bitvector, v2);
    DELELEMENT1(&possible_cycle_vertices, v2);

    contains_forbidden_induced_cycle_recursion(forbiddencyclesize);

    return contains_forbidden_cycle;
}

/******************************************************************************/

//Warning: 4-cycle is not necessarily induced!!!
static int
contains_fourcycle() {
    int v;
    for(v = 0; v < nv; v++) {
        setword current_vertex_bitvector = 0;
        ADDELEMENT1(&current_vertex_bitvector, v);
        int i, j;
        for(i = 0; i < adj[v] - 1; i++)
            for(j = i + 1; j < adj[v]; j++)
                if((nautyg[current_graph[v][i]] & nautyg[current_graph[v][j]]) != current_vertex_bitvector)
                    //I.e. there are 2 neighbours of nv - 1 which have more than 1 common neighbour
                    return 1;
    }
        
    return 0;
}


/******************************************************************************/

//Returns 1 if the graph contains a square containing v1 and v2
//It is assumed that v1 and v2 are neighbours in the square
//Warning: 4-cycle is not necessarily induced!!!
static int
contains_fourcycle_last_edge(unsigned char v1, unsigned char v2) {
    
    setword current_vertex_bitvector = 0;
    ADDELEMENT1(&current_vertex_bitvector, v1);
    int i;
    for(i = 0; i < adj[v1]; i++) {
        int nbr = current_graph[v1][i];
        if(nbr != v2
                && (nautyg[nbr] & nautyg[v2]) != current_vertex_bitvector)
            return 1;
    }
    
    return 0;
}

/******************************************************************************/

static int
contains_triangle() {
    int i;
    for(i = 0; i < nv; i++) {
        //set *gv = GRAPHROW1(nautyg, nv - 1, MAXM);
        set *gv = GRAPHROW1(nautyg, i, MAXM);
        int neighbour_b1 = -1;
        while((neighbour_b1 = nextelement1(gv, 1, neighbour_b1)) >= 0) {
            setword intersect_neigh_n_b1 = *GRAPHROW1(nautyg, neighbour_b1, MAXM) & *gv;
            if(intersect_neigh_n_b1 != 0)
                return 1;
        }
    }

    return 0;
}

/******************************************************************************/

//Returns 1 if the graph contains a triangle containing v1 and v2
static int
contains_triangle_last_edge(unsigned char v1, unsigned char v2) {
    return (nautyg[v1] & nautyg[v2]) != 0;
}

/******************************************************************************/

//Important: it is assumed that candidate_edges are valid!
static void
add_edge_in_all_possible_ways(int max_deg) {
    
/*
    int i, j;
    for(i = 0; i < nv; i++)
        if(adj[i] < max_deg)
            for(j = i + 1; j < nv; j++)
                if(adj[j] < max_deg && !ISELEMENT1(&nautyg[i], j)) {
                    //Add edge between i and j
                    add_edge(i, j);
                    
                    //extend
                    extend(max_deg);
                    
                    //remove edge
                    remove_edge(i, j);
                }
*/
    
    int i;
    for(i = 0; i < num_candidate_edges; i++) {
        //It is assumed that the resulting graph is still nonham and that the maxdeg is still ok!
        add_edge(candidate_edges[i][0], candidate_edges[i][1]);
        
        //i.e. now don't search for type A obstructions anymore
        extend(max_deg, 0, 0);
        
        remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
    }
    
}

/******************************************************************************/

//Important: it is assumed that candidate_edges are valid!
//Only performs expansions which increase the degree of deg two vertex!
static void
add_edge_in_all_possible_ways_given_endpoint(int max_deg, int given_endvertex) {

    int i;
    for(i = 0; i < num_candidate_edges; i++) {
        if(candidate_edges[i][0] == given_endvertex || candidate_edges[i][1] == given_endvertex) {
            //It is assumed that the resulting graph is still nonham and that the maxdeg is still ok!
            add_edge(candidate_edges[i][0], candidate_edges[i][1]);
            
            //i.e. now don't search for type A obstructions anymore
            extend(max_deg, 0, 0);

            remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
        }
    }
    
}

/******************************************************************************/

//Important: it is assumed that candidate_edges are valid!
//Only performs expansions which increase the degree of deg two vertex!
static void
add_edge_in_all_possible_ways_type_A_obstruction(int max_deg, int num_w_vertices, 
        setword w_bitvector) {

    //Copy since will be overwritten!
    //Is only valid for vertices in w_bitvector!
    int component_nr_of_vertex_local[nv];
    memcpy(component_nr_of_vertex_local, component_nr_of_vertex, sizeof(int) * nv);
    
    int i;
    for(i = 0; i < num_candidate_edges; i++) {
        //Try each adding each non-edge joining two components of w_vertices
        if(ISELEMENT1(&w_bitvector, candidate_edges[i][0]) 
                && ISELEMENT1(&w_bitvector, candidate_edges[i][1])
                //If girth 3: most efficient: must be in diff comp
                //If girth > 3: most efficient: at least one deg <= 1 but can be in same comp
                && ((girth < 4 && (component_nr_of_vertex_local[candidate_edges[i][0]] != component_nr_of_vertex_local[candidate_edges[i][1]]))
                    || (girth >= 4 && (POPC(nautyg[candidate_edges[i][0]] & w_bitvector) <= 1 || POPC(nautyg[candidate_edges[i][1]] & w_bitvector) <= 1)))
                
                
                //No must also consider the case where the both endpoints of the edge are in the same component!
                //&& component_nr_of_vertex_local[candidate_edges[i][0]] != component_nr_of_vertex_local[candidate_edges[i][1]]
                //Must add an edge where one of the endpoints has degree <= 1 in G[W] 
                //in order to be able to destroy the type A obstruction at some point!
                //(and can later then still add other edges)
                //&& (POPC(nautyg[candidate_edges[i][0]] & w_bitvector) <= 1
                //   || POPC(nautyg[candidate_edges[i][1]] & w_bitvector) <= 1)
                ) {
            //It is assumed that the resulting graph is still nonham and that the maxdeg is still ok!
            add_edge(candidate_edges[i][0], candidate_edges[i][1]);

            //The new W' is obtained from the old W by deleting every vertex which has deg 3 in <W>
            //Since then <W'> remains a set of disjoint paths
            
            //The only vertices which could have deg 3 in <W'> are candidate_edges[i][0] and candidate_edges[i][1]
            
            setword w_bitvector_new = w_bitvector;
            int num_w_vertices_new = num_w_vertices;
            
            int j;
            for(j = 0; j < 2; j++) {
                int num_nbrs_in_w_vertices = POPC(nautyg[candidate_edges[i][j]] & w_bitvector);
                
                if(num_nbrs_in_w_vertices > 3) {
                    fprintf(stderr, "Error: cannot have degree more than 3 in induced graph!\n");
                    exit(1);
                }

                if(num_nbrs_in_w_vertices == 3) {
                    DELELEMENT1(&w_bitvector_new, candidate_edges[i][j]);
                    num_w_vertices_new--;
                }
            }
            
            extend(max_deg, num_w_vertices_new, w_bitvector_new);

            remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
        }
    }
    
}

/******************************************************************************/

//Important: it is assumed that candidate_edges are valid!
//Important: it is assumed that vertices which are not in W have deg MAXN in degree_of_type_B_vertex[]
//Only performs expansions where at least one of the endpoints of the candidate edges is a vertex with deg <= 1
static void
add_edge_in_all_possible_ways_type_B_obstruction(int max_deg, int degree_of_type_B_vertex[]) {

    int i;
    for(i = 0; i < num_candidate_edges; i++) {
        if(degree_of_type_B_vertex[candidate_edges[i][0]] <= 1 
                || degree_of_type_B_vertex[candidate_edges[i][1]] <= 1) {
            //It is assumed that the resulting graph is still nonham and that the maxdeg is still ok!
            add_edge(candidate_edges[i][0], candidate_edges[i][1]);
            

            //i.e. now don't search for type A obstructions anymore
            extend(max_deg, 0, 0);

            remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
        }
    }
    
}

/******************************************************************************/

//Important: it is assumed that candidate_edges are valid!
static void
add_edge_in_all_possible_ways_type_C_obstruction(int max_deg, setword w_vertices_bitvector, 
        unsigned char vertex_v) {
    
    int num_neighbours_in_W[nv];
    int i;
    for(i = 0; i < nv; i++)
        if(!ISELEMENT1(&w_vertices_bitvector, i) && i != vertex_v)
            num_neighbours_in_W[i] = POPC(nautyg[i] & w_vertices_bitvector);
        else
            num_neighbours_in_W[i] = MAXN;
    
    for(i = 0; i < num_candidate_edges; i++) {
        //Only add edges for which condition 1 or 2 holds
        //1. Both endpoints of the edge are part of W
        //2. One endpoint of the edge is part of W and the other endpoint b is in X-v
        //   and b has at most 1 neighbour in W before adding the edge
        
        //Could be done more efficiently, but is no bottleneck
        
        int perform_expansion = 0;
        
        if(ISELEMENT1(&w_vertices_bitvector, candidate_edges[i][0]) 
                && ISELEMENT1(&w_vertices_bitvector, candidate_edges[i][1])) {
            //Both endpoints are in W
            perform_expansion = 1;     
        } else if(ISELEMENT1(&w_vertices_bitvector, candidate_edges[i][0])
                //&& !ISELEMENT1(&w_vertices_bitvector, candidate_edges[i][1])  //Is automatically enforced by the next condition
                && num_neighbours_in_W[candidate_edges[i][1]] <= 1) { //Will automatically be different from vertex_v!
            //candidate_edges[i][0] is in W and candidate_edges[i][1] is in X-v
            //and has at most 1 nbr in W
            perform_expansion = 1;
        } else if(ISELEMENT1(&w_vertices_bitvector, candidate_edges[i][1])
                && num_neighbours_in_W[candidate_edges[i][0]] <= 1) { //Will automatically be different from vertex_v!
            //candidate_edges[i][1] is in W and candidate_edges[i][0] is in X-v
            //and has at most 1 nbr in W
            perform_expansion = 1;
        }
        
        if(perform_expansion) {
            //It is assumed that the resulting graph is still nonham and that the maxdeg is still ok!
            add_edge(candidate_edges[i][0], candidate_edges[i][1]);
            
            //i.e. now don't search for type A obstructions anymore
            extend(max_deg, 0, 0);

            remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
        }
    }
    
}

/******************************************************************************/

//Important: it is assumed that the current graph is nonham!
static void
update_candidate_edges(int max_deg) {
    int num_candidate_edges_new = 0;
    int i;
    for(i = 0; i < nv; i++)
        num_edges_with_given_endpoint[i] = 0;

    if(!test_almost_hypoham_complete) {
        for(i = 0; i < num_candidate_edges; i++) {
            if(adj[candidate_edges[i][0]] < max_deg
                    && adj[candidate_edges[i][1]] < max_deg
                    && !ISELEMENT1(&nautyg[candidate_edges[i][0]], candidate_edges[i][1])) {
                //Add edge between i and j
                add_edge(candidate_edges[i][0], candidate_edges[i][1]);

                //Last edge helps a little bit (but is not really a bottleneck, certianly not for larger girth)

                int is_valid_edge = 1;
                if(test_if_trianglefree && contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                    //if(test_if_trianglefree && contains_triangle())
                    is_valid_edge = 0;
                else if(test_girth && girth > 4 && contains_fourcycle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                    //else if(test_girth && girth > 4 && contains_fourcycle())
                    is_valid_edge = 0;
                else if(test_girth && girth > 5 && contains_forbidden_induced_cycle_last_edge(candidate_edges[i][0], candidate_edges[i][1], 5))
                    is_valid_edge = 0;
                else if(test_girth && girth > 6 && contains_forbidden_induced_cycle_last_edge(candidate_edges[i][0], candidate_edges[i][1], 6))
                    is_valid_edge = 0;
                else if(test_girth && girth > 7 && contains_forbidden_induced_cycle_last_edge(candidate_edges[i][0], candidate_edges[i][1], 7))
                    is_valid_edge = 0;

                //Could also postpone triangle-free test is no bottleneck!
                //if(!test_if_trianglefree || !contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1])) {
                if(is_valid_edge) {
                    //if(!contains_hamiltonian_cycle()) {
                    //Add to candidate edges if result is nonham
                    candidate_edges[num_candidate_edges_new][0] = candidate_edges[i][0];
                    candidate_edges[num_candidate_edges_new][1] = candidate_edges[i][1];
                    num_candidate_edges_new++;
                    num_edges_with_given_endpoint[candidate_edges[i][0]]++;
                    num_edges_with_given_endpoint[candidate_edges[i][1]]++;

                    //    times_not_hamiltonian++;
                    //} //else
                    //    times_hamiltonian++;

                }

                //remove edge
                remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
            }
        }
    } else {
        int vertex_with_deg_larger_than_maxdeg = -1;
        int num_larger_than_max_deg = 0;

        for(i = 0; i < nv; i++)
            if(adj[i] > max_deg) {
                vertex_with_deg_larger_than_maxdeg = i;
                num_larger_than_max_deg++;
            }

        if(num_larger_than_max_deg > 1) {
            fprintf(stderr, "Error: too many larger than max deg!\n");
            exit(1);
        }

        if(vertex_with_deg_larger_than_maxdeg != -1) {
            for(i = 0; i < num_candidate_edges; i++) {
                if((adj[candidate_edges[i][0]] < max_deg || candidate_edges[i][0] == vertex_with_deg_larger_than_maxdeg)
                        && (adj[candidate_edges[i][1]] < max_deg || candidate_edges[i][1] == vertex_with_deg_larger_than_maxdeg)
                        && !ISELEMENT1(&nautyg[candidate_edges[i][0]], candidate_edges[i][1])) {
                    //Add edge between i and j
                    add_edge(candidate_edges[i][0], candidate_edges[i][1]);

                    //Last edge helps a little bit (but is not really a bottleneck, certianly not for larger girth)

                    int is_valid_edge = 1;
                    if(test_if_trianglefree && contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                        //if(test_if_trianglefree && contains_triangle())
                        is_valid_edge = 0;
                    else if(test_girth && girth > 4 && contains_fourcycle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                        //else if(test_girth && girth > 4 && contains_fourcycle())
                        is_valid_edge = 0;
                    else if(test_girth && girth > 5 && contains_forbidden_induced_cycle_last_edge(candidate_edges[i][0], candidate_edges[i][1], 5))
                        is_valid_edge = 0;

                    //Could also postpone triangle-free test is no bottleneck!
                    //if(!test_if_trianglefree || !contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1])) {
                    if(is_valid_edge) {
                        //if(!contains_hamiltonian_cycle()) {
                        //Add to candidate edges if result is nonham
                        candidate_edges[num_candidate_edges_new][0] = candidate_edges[i][0];
                        candidate_edges[num_candidate_edges_new][1] = candidate_edges[i][1];
                        num_candidate_edges_new++;
                        num_edges_with_given_endpoint[candidate_edges[i][0]]++;
                        num_edges_with_given_endpoint[candidate_edges[i][1]]++;

                        //    times_not_hamiltonian++;
                        //} //else
                        //    times_hamiltonian++;

                    }

                    //remove edge
                    remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
                }
            }
        } else {
            for(i = 0; i < num_candidate_edges; i++) {
                if((adj[candidate_edges[i][0]] < max_deg
                        || adj[candidate_edges[i][1]] < max_deg) //Not both can be larger than max deg!
                        && !ISELEMENT1(&nautyg[candidate_edges[i][0]], candidate_edges[i][1])) {
                    //Add edge between i and j
                    add_edge(candidate_edges[i][0], candidate_edges[i][1]);

                    //Last edge helps a little bit (but is not really a bottleneck, certianly not for larger girth)

                    int is_valid_edge = 1;
                    if(test_if_trianglefree && contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                        //if(test_if_trianglefree && contains_triangle())
                        is_valid_edge = 0;
                    else if(test_girth && girth > 4 && contains_fourcycle_last_edge(candidate_edges[i][0], candidate_edges[i][1]))
                        //else if(test_girth && girth > 4 && contains_fourcycle())
                        is_valid_edge = 0;
                    else if(test_girth && girth > 5 && contains_forbidden_induced_cycle_last_edge(candidate_edges[i][0], candidate_edges[i][1], 5))
                        is_valid_edge = 0;

                    //Could also postpone triangle-free test is no bottleneck!
                    //if(!test_if_trianglefree || !contains_triangle_last_edge(candidate_edges[i][0], candidate_edges[i][1])) {
                    if(is_valid_edge) {
                        //if(!contains_hamiltonian_cycle()) {
                        //Add to candidate edges if result is nonham
                        candidate_edges[num_candidate_edges_new][0] = candidate_edges[i][0];
                        candidate_edges[num_candidate_edges_new][1] = candidate_edges[i][1];
                        num_candidate_edges_new++;
                        num_edges_with_given_endpoint[candidate_edges[i][0]]++;
                        num_edges_with_given_endpoint[candidate_edges[i][1]]++;

                        //    times_not_hamiltonian++;
                        //} //else
                        //    times_hamiltonian++;

                    }

                    //remove edge
                    remove_edge(candidate_edges[i][0], candidate_edges[i][1]);
                }
            }
        }
    }
    
    num_candidate_edges = num_candidate_edges_new;
}

/******************************************************************************/

//Important: it is assumed that the current graph is nonham!
static void
compute_candidate_edges(int max_deg) {
    num_candidate_edges = 0;

    int i, j;
    if(!test_almost_hypoham_complete) {
        for(i = 0; i < nv; i++)
            if(adj[i] < max_deg)
                for(j = i + 1; j < nv; j++)
                    if(adj[j] < max_deg && !ISELEMENT1(&nautyg[i], j)) {
                        //Add edge between i and j
                        add_edge(i, j);

                        if(!contains_hamiltonian_cycle()) {
                            //Add to candidate edges if result is nonham
                            candidate_edges[num_candidate_edges][0] = i;
                            candidate_edges[num_candidate_edges][1] = j;
                            num_candidate_edges++;
                        }

                        //remove edge
                        remove_edge(i, j);
                    }
    } else {
        for(i = 0; i < nv; i++)
            //if(adj[i] < max_deg)
            for(j = i + 1; j < nv; j++)
                //if(adj[j] < max_deg && !ISELEMENT1(&nautyg[i], j)) {
                //Not both can become larger than max deg!
                if((adj[i] < max_deg || adj[j] < max_deg)
                        && !ISELEMENT1(&nautyg[i], j)) {
                    //Add edge between i and j
                    add_edge(i, j);

                    if(!contains_hamiltonian_cycle()) {
                        //Add to candidate edges if result is nonham
                        candidate_edges[num_candidate_edges][0] = i;
                        candidate_edges[num_candidate_edges][1] = j;
                        num_candidate_edges++;
                    }

                    //remove edge
                    remove_edge(i, j);
                }
    }
}

/******************************************************************************/

//Returns 1 if the graph contains a degree 2 vertex, else returns 0.
//If 1 is returned, best_degree_two_vertex points to the *best* deg 2 vertex
//Important: it is assumed that the graph has min deg 2
static int
contains_degree_two_vertex(int *best_degree_two_vertex) {
    int i;
    for(i = 0; i < nv; i++) //smallest label first is best
        if(adj[i] == 2) {
            *best_degree_two_vertex = i;
            return 1;
        }
    
    return 0;
}


/******************************************************************************/

//Returns 1 if the graph contains a degree 2 vertex, else returns 0.
//If 1 is returned, best_degree_two_vertex points to the *best* deg 2 vertex
//Important: it is assumed that the graph has min deg 2
static int
contains_degree_one_vertex(int *best_degree_two_vertex) {
    int i;
    for(i = 0; i < nv; i++) //smallest label first is best
        if(adj[i] == 1) {
            *best_degree_two_vertex = i;
            return 1;
        }
    
    return 0;
}

/******************************************************************************/

static int
vertex_is_part_of_triangle(int vertex) {
    int i;
    for(i = 0; i < adj[vertex]; i++)
        if((nautyg[current_graph[vertex][i]] & nautyg[vertex]) != 0)
            return 1;  
    
    return 0;
    
}

/******************************************************************************/

//Returns 1 if the graph contains a degree 2 vertex, else returns 0.
//If 1 is returned, best_degree_two_vertex points to the *best* deg 2 vertex
//Important: it is assumed that the graph has min deg 2
static int
contains_degree_three_vertex_which_is_part_of_triangle(int *best_triangle_deg_three_vertex) {
/*
    int current_best_deg_three_vertex;
    int min_num_cand_edges = MAXN;
    int i;
    for(i = 0; i < nv; i++)
        if(adj[i] == 3 && vertex_is_part_of_triangle(i))
            //Is very helpful!
            //Largest label seems to be even better
            if(num_edges_with_given_endpoint[i] <= min_num_cand_edges) {
                min_num_cand_edges = num_edges_with_given_endpoint[i];
                current_best_deg_three_vertex = i;
            }
    
    //fprintf(stderr, "min_num_cand_edges: %d\n", min_num_cand_edges);
    
    if(min_num_cand_edges < MAXN) {
        *best_triangle_deg_three_vertex = current_best_deg_three_vertex;
        return 1;
    } else
        return 0;    
*/
    
    int i;
    for(i = 0; i < nv; i++) //Min or max label doesn't make much difference
        if(adj[i] == 3 && vertex_is_part_of_triangle(i)) {
            *best_triangle_deg_three_vertex = i;
            return 1;
        }
    
    return 0;
}

/******************************************************************************/

/**
 * DFS algorithm which recursively marks all vertices which can be reached from
 * current_vertex.
 * Complexity: O(|V|+|E|)
 */
static void mark_dfs(unsigned char current_vertex, setword w_bitvector) {
    MARK(current_vertex);
    component_nr_of_vertex[current_vertex] = current_type_A_component;
    
    int i;
    for(i = 0; i < adj[current_vertex]; i++)
        if(!ISMARKED(current_graph[current_vertex][i]) && ISELEMENT1(&w_bitvector, current_graph[current_vertex][i]))
            mark_dfs(current_graph[current_vertex][i], w_bitvector);
    
}

/******************************************************************************/

static int is_type_A_obstruction(int num_w_vertices, setword w_bitvector) {
    if(num_w_vertices == 0) {
        fprintf(stderr, "Error: num_w_vertices should be > 0\n");
        exit(1);
    }

    //Compute components
    //It is assumed that G[W] will always be a set of disjoint paths, so this is easy to compute    
    
    RESETMARKS;
    
    current_type_A_component = 0;
    
    int i;
    for(i = 0; i < nv; i++)
        if(!ISMARKED(i) && ISELEMENT1(&w_bitvector, i)) {
            //component_nr_of_vertex[i] = current_type_A_component;
            mark_dfs(i, w_bitvector);
            current_type_A_component++;
        }
    
    //fprintf(stderr, "num components: %d\n", num_type_A_components);
    
    if((!test_platypus && current_type_A_component >= (nv - num_w_vertices))
            || (test_platypus && current_type_A_component > (nv - num_w_vertices))) {
        
        if(nv - num_w_vertices < 2) {
            fprintf(stderr, "Error: set X should have size at least 2!\n");
            exit(1);
        }
        
        return 1;
    } else
        return 0;
}

/******************************************************************************/

int compute_k_function(unsigned char current_set[], int current_set_size, 
        setword current_set_bitvector) {
    if(current_set_size == 0)
        return 0;

    //P_POPCOUNT not useful since deg is rarely 0!
    int i;
    for(i = 0; i < current_set_size; i++)
        degree_in_graph_induced_by_partition[current_set[i]] = POPC(nautyg[current_set[i]] & current_set_bitvector);
    
    int k_value = 0;
    
    int num_remaining_vertices = current_set_size;
    
    //First count number of isolated vertices
    for(i = 0; i < current_set_size; i++)
        if(degree_in_graph_induced_by_partition[current_set[i]] == 0) {
            k_value++;
            num_remaining_vertices--;
        }
    
    //Now count number of isolated edges
    for(i = 0; i < current_set_size; i++)
        if(degree_in_graph_induced_by_partition[current_set[i]] == 1) {
            int j;
            for(j = 0; j < adj[current_set[i]]; j++) {
                int neighbour = current_graph[current_set[i]][j];
                if(i < neighbour && ISELEMENT1(&current_set_bitvector, neighbour)
                        && degree_in_graph_induced_by_partition[neighbour] == 1) {
                    //Isolated edge found
                    k_value++;

                    //To avoid that this vertex is counted twice!
                    degree_in_graph_induced_by_partition[current_set[i]] = 0;
                    degree_in_graph_induced_by_partition[neighbour] = 0;

                    num_remaining_vertices -= 2;
                }
            }
        }
    
    int rest = 0;
    
    if(num_remaining_vertices < 0) {
        fprintf(stderr, "Error: num_remaining_vertices < 0\n");
        exit(1);
    }    
    
    //Now count number of remaining deg 1 vertices (at least if the rest of the graph isn't empty)
    if(num_remaining_vertices > 0) {
        int num_remaining_deg_1_vertices = 0;
        for(i = 0; i < current_set_size; i++)
            //Isolated edges now have degree 0
            if(degree_in_graph_induced_by_partition[current_set[i]] == 1)
                num_remaining_deg_1_vertices++;

        //Ceil(a/2) = Floor((a+1)/2)
        rest = MAX(1, (num_remaining_deg_1_vertices + 1) / 2);
    }
    
    return k_value + rest;
}

/******************************************************************************/

static void
generate_all_possible_partitions(unsigned char current_set[], int current_set_size, 
        setword current_set_bitvector, int next_vertex) {
    if(type_B_obstruction_found)
        return;
    
    if(current_set_size > nv - 2)
        return; //Remaining set must contain at least 2 vertices!
    
    if(current_set_size > 0) {
        //Test if is type B obstruction
        
        int k_value = compute_k_function(current_set, current_set_size, current_set_bitvector);
        if((!test_platypus && k_value >= (nv - current_set_size))
                || (test_platypus && k_value > (nv - current_set_size))) {
            
            //TODO: later search for best type B vertices?!
            
            type_B_obstruction_found = 1;
            type_B_W_vertices = current_set_bitvector;
            return;
        }
            
    }
    
    int i;
    for(i = next_vertex; i < nv; i++) {
        current_set[current_set_size] = i;
        setword current_set_bitvector_new = current_set_bitvector;
        ADDELEMENT1(&current_set_bitvector_new, i);
        generate_all_possible_partitions(current_set, current_set_size + 1, 
            current_set_bitvector_new, i + 1);
    }
}

/******************************************************************************/

//Returns 1 if the graph contains a degree 2 vertex, else returns 0.
//If 1 is returned, best_degree_two_vertex points to the *best* deg 2 vertex
//Important: it is assumed that the graph has min deg 2
static int
contains_type_B_obstruction(int degree_of_type_B_vertex[]) {
    
    //TODO: Look at all independent sets
    //W = indep set + vertices which are not adjacent to it
    
    //Compute k_function and test if it is a type B obstruction
    type_B_obstruction_found = 0;
    
    unsigned char current_set[nv];
    setword current_set_bitvector = 0;
    generate_all_possible_partitions(current_set, 0, current_set_bitvector, 0);
    
    if(type_B_obstruction_found) {
        int i;
        for(i = 0; i < nv; i++)
            if(ISELEMENT1(&type_B_W_vertices, i))
                degree_of_type_B_vertex[i] = degree_in_graph_induced_by_partition[i];
            else
                degree_of_type_B_vertex[i] = MAXN;
        return 1;
    } else
        return 0;
    
}


/**************************************************************************/

setword w_vertices;
unsigned char removed_vertex;

static void
search_all_independent_sets(unsigned char current_set[], int current_set_size, setword current_set_bitvector,
        ADJACENCY vertices, int next_index, setword possible_vertices, int num_possible_vertices, int order) {

    if(type_C_obstruction_found)
        return;
    
    if(current_set_size == order) {

        if(current_set_size > nv - 2) {
            fprintf(stderr, "Error: remaining set must contain at least 2 vertices!\n");
            exit(1);
        }
        
        if(order < 1) {
            fprintf(stderr, "Error: order must be at least 1\n");
            exit(1);
        }
        
        //For every vertex in X
        //Compute number of neighbours in W
        
        int num_with_one_nbr_in_W = 0;
        int num_with_more_than_one_nbr_in_W = 0;
        int num_nbrs_in_W[nv];
        int i;
        //for(i = 0; i < nv; i++)
        //    num_nbrs_in_W[i] = 0;
        for(i = 0; i < nv; i++)
            if(!ISELEMENT1(&current_set_bitvector, i)) {
                num_nbrs_in_W[i] = POPC(nautyg[i] & current_set_bitvector);
                if(num_nbrs_in_W[i] == 1)
                    num_with_one_nbr_in_W++;
                else if(num_nbrs_in_W[i] > 1)
                    num_with_more_than_one_nbr_in_W++;
            }
        
        //Nu kijken of er een vertex v is die verwijderd kan worden...
        //Wil natuurlijk altijd een vertex met zoveel mogelijk buren in W verwijderen!
        
        //Hoe beste v kiezen? Bv met min of max deg? Speelt wellicht geen rol...
        

        //TODO: hoeft niet te doen als 
        //2 * num_with_more_than_one_nbr_in_W + num_with_one_nbr_in_W - 2 >= 2 * current_set_size
        //Maar is wellicht geen bottleneck?
        for(i = 0; i < nv; i++)
            if(!ISELEMENT1(&current_set_bitvector, i)) {
                int num_to_substract = 0;
                if(num_nbrs_in_W[i] == 1)
                    num_to_substract = 1;
                else if(num_nbrs_in_W[i] > 1)
                    num_to_substract = 2;

                if((!test_platypus && 2 * num_with_more_than_one_nbr_in_W + num_with_one_nbr_in_W - num_to_substract < 2 * current_set_size)
                        || (test_platypus && 2 * num_with_more_than_one_nbr_in_W + num_with_one_nbr_in_W - num_to_substract < 2 * current_set_size - 2)) {
                    //Type C obstruction found
                    
                    //Store w bitvector and v
                    w_vertices = current_set_bitvector;
                    removed_vertex = i;

                    type_C_obstruction_found = 1;

                    return;

                }
            }
        
        number_of_indep_sets_found[order]++;
        
        return;
    }

    int i;
    unsigned char vertex;
    setword current_set_bitvector_new;
    setword possible_vertices_new;
    int num_possible_vertices_new;
    int num_required = order - current_set_size;
    //int num_required = 1;
    //Does not happen that i >= max_size!
    //for(i = next_index; (i < max_size) && num_possible_vertices >= num_required &&  !(*is_subgraph) ; i++) {
    for(i = next_index; (num_possible_vertices >= num_required); i++) {
        //for(i = next_index; i < number_of_vertices; i++) {
        vertex = vertices[i];
        if(ISELEMENT1(&possible_vertices, vertex)) {
            num_possible_vertices--;
            DELELEMENT1(&possible_vertices, vertex);
            //Neighbours of vertex are also forbidden since we can't have any more edges
            int num_extra_forbidden = POPC(possible_vertices & nautyg[vertex]);
            num_possible_vertices_new = num_possible_vertices - num_extra_forbidden;
            if(num_extra_forbidden > 0)
                possible_vertices_new = possible_vertices & ~nautyg[vertex];
            else
                possible_vertices_new = possible_vertices;

                current_set_bitvector_new = current_set_bitvector;
                ADDELEMENT1(&current_set_bitvector_new, vertex);
            
            current_set[current_set_size] = vertex;

            search_all_independent_sets(current_set, current_set_size + 1,
                    current_set_bitvector_new, vertices, i + 1, possible_vertices_new,
                    num_possible_vertices_new, order);
        }
    }
}

/******************************************************************************/

static int
contains_type_C_obstruction(setword *w_vertices_bitvector, unsigned char *vertex_v) {

    unsigned char vertices[nv];
    int i;
    for(i = 0; i < nv; i++)
        vertices[i] = i;
    setword possible_vertices = ALLMASK(nv);
    //number_of_indep_sets_found = 0;
    //unsigned char current_set[MAXN];
    unsigned char current_set[nv];
    
    type_C_obstruction_found = 0;
    
    //Computing the indep sets is not a bottleneck!
    //nv - 2 since remaining set must contain at least 2 vertices
    //Seems best to have IS as small as possible (such that there are few possible expansions)
    //But experiments show that large IS's first is a lot faster!
    //for(i = 2; i <= nv - 2; i++) {
    for(i = nv - 2; i >= 2; i--) {
        //Search all indep sets of size i
        number_of_indep_sets_found[i] = 0;
        search_all_independent_sets(current_set, 0, 0, vertices, 0, possible_vertices,
                nv, i);
        //fprintf(stderr, "Number of indep sets of order %d: %d\n", i, number_of_indep_sets_found[i]);

        if(type_C_obstruction_found) {

            //TODO: store vertices in W, X-v and vertex v
            //Of gewoon W en v geven, is voldoende!
            
            *w_vertices_bitvector = w_vertices;
            *vertex_v = removed_vertex;

            //fprintf(stderr, "type C obstr found (IS set size: %d)\n", POPC(*w_vertices_bitvector));
            //exit(1);
            
            return 1;
        }        
    }        
    
    
    return 0;
    
}


/******************************************************************************/

/**
 * Returns the number of edges of the current graph.
 */
static int
determine_number_of_edges() {
    int ne = 0;
    int i;
    for(i = 0; i < nv; i++)
        ne += adj[i];
    
    return ne / 2;
}

/******************************************************************************/

/**
 * Returns 1 if the current graph is planar, else returns 0.
 * 
 * Uses planarity.c - code for planarity testing of undirected graphs by
 * Boyer and Myrvold.
 */
static int
is_planar() {
    int ne = determine_number_of_edges();
    //Never happens (because of previous_extension_was_not_colourable = 1)
    //if(ne > (3*nv - 6)) {
    //    fprintf(stderr, "cant be planar\n");
    //    return 0; //Can't be planar in this case
    //}
    
    sparsegraph sg;
    SG_INIT(sg);
    SG_ALLOC(sg, nv, 2*ne, "malloc");    
    nauty_to_sg(nautyg, &sg, 1, nv);
    
/*
    printgraph(current_graph, adj, nv);
    print_sparse_graph_nauty(sg);
    fprintf(stderr, "\n");
*/
    
/*
    DYNALLSTAT(t_ver_sparse_rep,V,V_sz);
    DYNALLSTAT(t_adjl_sparse_rep,A,A_sz);
    DYNALLOC1(t_ver_sparse_rep, V, V_sz, nv, "planarg");
    DYNALLOC1(t_adjl_sparse_rep, A, A_sz, 2 * ne + 1, "planarg");
*/
    t_ver_sparse_rep V[nv];
    t_adjl_sparse_rep A[2*ne + 1];
    
    int k = 0;
    int i, j;
    for(i = 0; i < nv; ++i)
        if(sg.d[i] == 0)
            V[i].first_edge = NIL;
        else {
            V[i].first_edge = k;
            for(j = sg.v[i]; j < sg.v[i] + sg.d[i]; ++j) {
                A[k].end_vertex = sg.e[j];
                A[k].next = k + 1;
                ++k;
            }
            A[k - 1].next = NIL;
        }

    if(k != 2 * ne) {
        fprintf(stderr, "Error: decoding error (planarg)\n");
        exit(1);
    }
    
    SG_FREE(sg);

    int nbr_c;
    int edge_pos, v, w;
    t_dlcl **dfs_tree, **back_edges, **mult_edges;
    t_ver_edge *embed_graph;
    
    //TODO: possible optimisation: only start from last vertex?
    int ans = sparseg_adjl_is_planar(V, nv, A, &nbr_c,
                                 &dfs_tree, &back_edges, &mult_edges,
                                 &embed_graph, &edge_pos, &v, &w);
    
    
    //Don't forget this otherwise out of mem!
    sparseg_dlcl_delete(dfs_tree, nv);
    sparseg_dlcl_delete(back_edges, nv);
    sparseg_dlcl_delete(mult_edges, nv);
    embedg_VES_delete(embed_graph, nv);
    
    return ans;
    
    
}

/******************************************************************************/

//removed_vertex: counts from 0
void copy_graph_to_minibaum_format_removed_vertex(unsigned char removed_vertex) {

    current_graph_mb[0][0] = nv - 1;
    
    int i;    
    for(i = 0; i < nv; i++) {
        if(i != removed_vertex) {
            int offset_one = 0;
            if(i > removed_vertex)
                offset_one++;
            //i is mapped to i - offset_one
            adj_mb[i - offset_one + 1] = 0;
            int j;
            for(j = 0; j < adj[i]; j++) {
                int nbr = current_graph[i][j];
                if(nbr != removed_vertex) {
                    int offset_nbr = 0;
                    if(nbr > removed_vertex)
                        offset_nbr++;

                    current_graph_mb[i - offset_one + 1][adj_mb[i - offset_one + 1]++] = nbr - offset_nbr + 1;
                }
            }
            current_graph_mb[i - offset_one + 1][adj_mb[i - offset_one + 1]] = leer;
        }
    }    
    
    //printgraph_minibaum(current_graph_mb, adj_mb, nv);    
}

/******************************************************************************/

//Returns 1 if the removal of the exceptional vertex yields a ham graph
//If this is the case, this graph cannot be (or lead to) an almost hypoham graph
//Note: the except vertex is a vertex with deg > max_deg (if such a vertex is present)
static int removal_of_exceptional_vertex_yields_ham_graph(int max_deg) {
    
    int vertex_with_deg_larger_than_maxdeg;
    int num_larger_than_max_deg = 0;
    int i;
    for(i = 0; i < nv; i++)
        if(adj[i] > max_deg) {
            vertex_with_deg_larger_than_maxdeg = i;
            num_larger_than_max_deg++;
        }

    if(num_larger_than_max_deg > 1) {
        fprintf(stderr, "Error: too many larger than max deg!\n");
        exit(1);
    }    
    
    if(num_larger_than_max_deg == 0)
        return 0;
    
    copy_graph_to_minibaum_format_removed_vertex(vertex_with_deg_larger_than_maxdeg);
    return hammi_start(current_graph_mb, adj_mb);
    
    //The method below is much slower

/*
    RESETMARKS;
    MARK(vertex_with_deg_larger_than_maxdeg); //i is forbidden

    //Best to start with vertex of min deg!
    //But won't make much difference since this is not a bottleneck!
    int startvertex = (vertex_with_deg_larger_than_maxdeg + 1) % nv;

    current_path[0] = startvertex;
    current_pathsize = 1;
    MARK(startvertex);

    cycle_found = 0;
    search_cycle(nv - 1);
    
    return cycle_found;
*/
    
}

/******************************************************************************/

//If num_w_vertices == 0, it doesn't test for type A obstructions
//Once a specific type A obstruction is destroyed, it will remain destroyed
static void extend(int max_deg, int num_w_vertices, setword w_bitvector) {
    //Can now happen since starting from multiple startgraphs
    if(nv != maxnv) {
        fprintf(stderr, "Error: wrong nv!\n");
        exit(1);
    }

    //Already testing this before when updating cand edges! This is not a bottleneck!
/*
    if(test_if_trianglefree) {
        if(contains_triangle()) {
            times_not_trianglefree++;

            return;
        } else
            times_trianglefree++;
    }
*/
    
    //Already testing this before when updating cand edges! This is not a bottleneck!
/*
    if(test_girth && girth > 4) {
        if(countains_fourcycle()) {
            //fprintf(stderr, "prune contains fourcycle!\n");
            
            return;
        }
        if(girth > 5) {
            fprintf(stderr, "Error: max supported girth is 5 for now\n");
            exit(1);
        }
    }    
*/
    
/*
    if(test_girth && girth > 5) {
        //It is assumed that graphs with triangles or squares were already filtered!
        //Moet eigenlijk enkel zoeken naar cykels die laatste boog bevatten!
        //This is a (small) bottleneck, so doing this earlier!
        if(contains_forbidden_induced_cycle(5)) {
            //fprintf(stderr, "prune contains C5!\n");
            return;
        }
        if(girth > 6) {
            fprintf(stderr, "Error: max supported girth is 6 for now\n");
            exit(1);
        }
    }        
*/
    
/*
    if(test_girth && girth > 6) {
        //Becomes bottleneck for large nv, so doing this earlier!
        if(contains_forbidden_induced_cycle(6)) {
            return;
        }
        if(girth > 7 && contains_forbidden_induced_cycle(7)) {
            return;
        }        
        if(girth > 8) {
            fprintf(stderr, "Error: max supported girth is 8 for now\n");
            exit(1);
        }
    }     
*/
    
    //If parent didn't contain a ham path, then G+e is certainly nonhamiltonian
    if(contains_hamiltonian_cycle()) {
        times_hamiltonian++;
        return;
    } else
        times_not_hamiltonian++;

    //Only doing this after calling nauty is not any faster or slower
    //Is because usually exceptional vertex not discernible?
    if(test_almost_hypoham_complete 
            && removal_of_exceptional_vertex_yields_ham_graph(max_deg)) {
        times_except_vertex_ham++;
        return;
    } else
        times_except_vertex_not_ham++;        
    
    //Best to do ham test before planarity test!
    if(test_if_planar) {
        if(!is_planar()) {
            times_not_planar++;
            return;
        } else {
            times_planar++;
        }
    }        
    
    //malloc is not faster, but it uses less memory than using MAXN!
    graph *can_form = (graph *) malloc(sizeof (graph) * nv);
    if(can_form == NULL) {
        fprintf(stderr, "Error: can't get more space\n");
        exit(1);
    }
    //Don't forget +3 for triangle gadget!
    nauty(nautyg, lab, ptn, NULL, orbits, &options, &stats, workspace, WORKSIZE, MAXM, nv, can_form);
    //memcpy(can_form, nautyg_canon, sizeof (graph) * nv);

    int is_new_node;
    splay_insert(&list_of_worklists[max_deg], can_form, &is_new_node);
    //No, this is not allowed for almost hypoham, otherwise graphs would be pruned even though not all expansions would be performed due to exceptional vertex!
    //splay_insert(&list_of_worklists[0], can_form, &is_new_node);    
    if(!is_new_node) {
        num_graphs_generated_iso++;
        free(can_form); //Don't forget to free!
        return;
    } else //i.e. is canon
        num_graphs_generated_noniso++;
    
    if(num_edges_added > max_num_edges_added)
        max_num_edges_added = num_edges_added;
    
    if(num_edges_added == splitlevel) {
        splitcounter++;
        
        if(splitcounter == modulo)
            splitcounter = 0;
        
        if(splitcounter != rest)
            return;
    }        
    
    //Ok, now we know the graph is hamiltonian and not generated before
    //So now update the candidate edges!
    
    //Backup candidate edges
    EDGE candidate_edges_backup[num_candidate_edges];
    int num_candidate_edges_backup = num_candidate_edges;
    memcpy(candidate_edges_backup, candidate_edges, sizeof(EDGE) * num_candidate_edges);
    
    //fprintf(stderr, "num cand edges: %d\n", num_candidate_edges);
    
    //Update candidate edges
    //If G contains no ham path, then G+e cannot be hamiltonian
    //If this graph doesn't contain a ham path, then G+e is certainly nonhamiltonian
    //graph_is_certainly_nonham = !contains_hamiltonian_path();
    update_candidate_edges(max_deg);

    //if(num_candidate_edges > 0) { //Doesn't really help since this is no bottleneck
    //Also if not allowed since if no cand edges it could still be that this graph is hypoham!

    //If num_w_vertices == 0, it doesn't test for type A obstructions
    //Once a specific type A obstruction is destroyed, it will remain destroyed

    //Best only to test for type A obstructions if girth = 3 or girth = 4 and nonplanar
    if((girth < 4 || (girth == 4 && !test_if_planar)) && num_w_vertices > 0 && is_type_A_obstruction(num_w_vertices, w_bitvector)) {
        times_contains_type_A_obstruction++;

        //destroy in all possible ways
        add_edge_in_all_possible_ways_type_A_obstruction(max_deg, num_w_vertices, w_bitvector);

    } else {
        times_doesnt_contain_type_A_obstruction++;

        int best_vertex;
        if((!test_platypus && contains_degree_two_vertex(&best_vertex))
                || (test_platypus && contains_degree_one_vertex(&best_vertex))) {
            times_contains_deg_two_vertex++;
            //Apply all possible extensions which increase degree of this vertex

            add_edge_in_all_possible_ways_given_endpoint(max_deg, best_vertex);

        } else {
            times_doesnt_contain_deg_two_vertex++;

            setword w_vertices_bitvector;
            unsigned char vertex_v;

            //The hope is that type C obstructions are more restrictive than type B obstructions
            //if(!test_almost_hypoham_special && contains_type_C_obstruction(&w_vertices_bitvector, &vertex_v)) {
            if(contains_type_C_obstruction(&w_vertices_bitvector, &vertex_v)) {
                times_contains_type_C_obstruction++;

                add_edge_in_all_possible_ways_type_C_obstruction(max_deg, w_vertices_bitvector, vertex_v);

            } else {
                times_doesnt_contain_type_C_obstruction++;

                if(!test_platypus && !test_if_trianglefree && contains_degree_three_vertex_which_is_part_of_triangle(&best_vertex)) {
                    times_contains_deg_three_triangle_vertex++;

                    add_edge_in_all_possible_ways_given_endpoint(max_deg, best_vertex);

                } else {
                    times_doesnt_contain_deg_three_triangle_vertex++;

                    int degree_of_best_type_B_vertices[nv];
                    //if(!test_almost_hypoham_special && contains_type_B_obstruction(degree_of_best_type_B_vertices)) {
                    if(contains_type_B_obstruction(degree_of_best_type_B_vertices)) {
                        times_contains_type_B_obstruction++;
                        
                        add_edge_in_all_possible_ways_type_B_obstruction(max_deg, degree_of_best_type_B_vertices);

                    } else {
                        times_doesnt_contain_type_B_obstruction++;

                        times_no_lemmas_applicable++;

                        //If the graph contains an obstruction, it cannot be hypohamiltonian

                        if(!test_almost_hypoham_special) {

                            if(!test_platypus) {
                                //This is no bottleneck
                                if(is_hypohamiltonian()) {
                                    output_graph(current_graph, adj, nv);
                                }
                            } else {
                                if(is_hypotraceable()) {
                                    output_graph(current_graph, adj, nv);
                                }
                            }
                        } else {
                            if(is_almost_hypohamiltonian()) {
                                    output_graph(current_graph, adj, nv);
                                }
                        }

                        //Now apply all possible expansions
                        //Such that the max deg doesn't increase and such that the result remains nonham
                        add_edge_in_all_possible_ways(max_deg);
                    }
                }
            }
        }
    }

    //Restore candidate edges
    num_candidate_edges = num_candidate_edges_backup;
    memcpy(candidate_edges, candidate_edges_backup, sizeof (EDGE) * num_candidate_edges_backup);


}

/******************************************************************************/

//It is assumed that the current graph is a cycle with max_deg vertices
static void
init_construction_hypoham(unsigned char current_set[], int current_set_size,
        int next_vertex, int max_deg) {
    
    if(current_set_size > 0) {
        expand_vertex_set_no_remove(current_set, current_set_size);

        //Ham test first makes no difference!
        //int is_ham = contains_hamiltonian_cycle();
        //Girth test first helps a lot!
        //Especially for large girth!

        int perform_extension = 1;
        if(test_if_trianglefree && contains_triangle())
            perform_extension = 0;
        else if(test_girth && girth > 4 && contains_fourcycle())
            perform_extension = 0;
        else if(test_girth && girth > 5 && contains_forbidden_induced_cycle(5))
            perform_extension = 0;
        else if(test_girth && girth > 6 && contains_forbidden_induced_cycle(6))
            perform_extension = 0;
        else if(test_girth && girth > 7 && contains_forbidden_induced_cycle(7))
            perform_extension = 0;               
        
        //reduce
        reduce_vertex_set_no_remove(current_set, current_set_size);
        
        if(!perform_extension) {
            //If this set already contains a cycle of size < girth, it will still 
            //have a cycle of size < girth when more edges are added, so stop here
            //This is certainly very helpful when generating graphs with girth >= 7
            return;
        }
    }
    
    if(current_set_size == max_deg) {
        
        if(current_set_size < 1) {
            fprintf(stderr, "Error: current set size must be large enough!\n");
            exit(1);
        }

        //extend
        expand_vertex_set_no_remove(current_set, current_set_size);

        //Remark: could test for hamiltonicity sooner (i.e. f a graph is ham, it will remain ham if more edges are added)
        //But this not a bottleneck at all in init_construction_hypoham()
        
        //Test if is nonham
        if(!contains_hamiltonian_cycle()) {
            compute_candidate_edges(max_deg);
            
            //Initially W = V(G) \ {N(hub) \cup {hub}}
            //No W = V(G) \ N(hub)
            
            //printgraph(current_graph, adj, nv);
            
            //setword w_bitvector = ALLMASK(nv - 1);
            setword w_bitvector = ALLMASK(nv);
            w_bitvector &= ~nautyg[nv - 1];
            
            int num_w_vertices = POPC(w_bitvector);
            
            //fprintf(stderr, "w_bitvector: %d\n", w_bitvector);
            //fprintf(stderr, "num_w_vertices: %d\n", num_w_vertices);
            
            num_edges_added = 0;
            
            //Actually this was already tested, but is no bottleneck!
            int perform_extension = 1;
            if(test_if_trianglefree && contains_triangle())
                perform_extension = 0;
            else if(test_girth && girth > 4 && contains_fourcycle())
                perform_extension = 0;
            else if(test_girth && girth > 5 && contains_forbidden_induced_cycle(5))
                perform_extension = 0;            
            else if(test_girth && girth > 6 && contains_forbidden_induced_cycle(6))
                perform_extension = 0;            
            else if(test_girth && girth > 7 && contains_forbidden_induced_cycle(7))
                perform_extension = 0;                        
            
            if(perform_extension)
                extend(max_deg, num_w_vertices, w_bitvector); //Only extend nonham graphs
        }

        //reduce
        reduce_vertex_set_no_remove(current_set, current_set_size);

    } else {
        
        if(current_set_size > max_deg) {
            fprintf(stderr, "Error: curr set size is too big!\n");
            exit(1);
        }

        //Make other possible sets
        int i;
        for(i = next_vertex; i < nv; i++) {
            current_set[current_set_size] = i;

            init_construction_hypoham(current_set, current_set_size + 1,
                    i + 1, max_deg);

        }

    }

}

/******************************************************************************/

//Makes a cycle of a given size
static void
make_cycle(int cycle_size) {
    nv = cycle_size;
    int i;
    for(i = 0; i < cycle_size; i++) {
        adj[i] = 2;
        current_graph[i][0] = (i + cycle_size - 1) % cycle_size;
        current_graph[i][1] = (i + 1) % cycle_size;
    }
}

/******************************************************************************/

//Makes a path of a given size
static void
make_path(int cycle_size) {
    nv = cycle_size;
    
    adj[0] = 1;
    current_graph[0][0] = 1;
    
    int i;
    //for(i = 0; i < cycle_size; i++) {
    for(i = 1; i < cycle_size - 1; i++) {
        adj[i] = 2;
        current_graph[i][0] = (i + cycle_size - 1) % cycle_size;
        current_graph[i][1] = (i + 1) % cycle_size;
    }

    adj[i] = 1;
    current_graph[i][0] = (i + cycle_size - 1) % cycle_size;

}

/******************************************************************************/

static void
init_construction() {
    //options.userautomproc = save_generators;
    options.defaultptn = TRUE;
    //options.getcanon = FALSE;
    options.getcanon = TRUE;
    
    //Create a cycle with maxnv - 1 vertices
    if(!test_platypus)
        make_cycle(maxnv - 1);
    else
        make_path(maxnv - 1); 
    copy_nauty_graph(current_graph, adj, nv, nautyg);
    
    //printgraph(current_graph, adj, nv);
    //printgraph_nauty(nautyg, nv);    
    
    
    //Min deg is 3, since hypoham graphs must be 3-connected
    //There will be a limit on the maxdeg, but this probably not a problem since
    //such graphs will most likely rapidly lead to a hamiltonian graph?
    
    unsigned char current_set[MAXN];
    
    int max = nv;
    if(impose_max_deg)
        max = max_deg + 1;
    
    int start = 3;    
    if(test_platypus)
        start = 2; //Mindeg >= 2 for platypusses
    if(impose_min_max_deg)
        start = min_max_deg;
    int i;
    for(i = start; i < max; i++) {
        struct tms TMS;
        times(&TMS);
        unsigned int savetime_before = (unsigned int) TMS.tms_utime;
        
        unsigned long long int num_graphs_generated_noniso_before = num_graphs_generated_noniso;
        fprintf(stderr, "Generating graphs with max deg = %d\n", i);
        init_construction_hypoham(current_set, 0, 0, i);
        fprintf(stderr, "Num of noniso graphs generated: %llu\n", num_graphs_generated_noniso - num_graphs_generated_noniso_before);
        
        times(&TMS);
        unsigned int savetime = (unsigned int) TMS.tms_utime;
        fprintf(stderr, "CPU time: %.1f seconds.\n", (double) ((savetime - savetime_before) / time_factor));
        fprintf(stderr, "---\n");
        
        //Now free all canon forms in the tree with node list_of_worklists[i]
        //(As there is no method to clean an entire tree, but only the  canon forms take significant memory)
        splay_scan(list_of_worklists[i]);
    }

}

/******************************************************************************/

static void
initialize_splitting() {
    if(modulo == 1)
        splitlevel = 0;
    else {
        //TODO: could decrease splitlevel planar or girth case
        splitlevel = DESIRED_SPLITLEVEL;
        
        fprintf(stderr, "Splitlevel is %d (num edges added)\n", splitlevel);
    }
}

/******************************************************************************/

static void
perform_wordsize_checks() {
    /* Checks to test if the WORDSIZE and sizeof setwords is valid */
    if(WORDSIZE != 32 && WORDSIZE != 64) {
        fprintf(stderr, "Error: invalid value for wordsize: %d\n", WORDSIZE);
        fprintf(stderr, "Valid values are 32 and 64\n");
        exit(1);
    }

    if(MAXN != WORDSIZE) {
        fprintf(stderr, "Error: MAXN is not equal to WORDSIZE: %d vs %d\n", MAXN, WORDSIZE);
        fprintf(stderr, "Please compile with option -DMAXN=WORDSIZE\n");
        exit(1);
    }
    
    if(MAXM != 1) {
        fprintf(stderr, "Error: expected MAXM 1 instead of %d\n", MAXM);
        exit(1);
    }

    if(WORDSIZE == 32) {
        fprintf(stderr, "32 bit mode\n");
        if(sizeof (unsigned int) < 4) {
            fprintf(stderr, "Error: unsigned ints should be at least 32 bit -- sorry.\n");
            exit(1);
        }
        if(sizeof (setword) < 4) {
            fprintf(stderr, "Error: this version relies on 32 bit setwords -- sorry.\n");
            exit(1);
        }
        if(sizeof (setword) != 4) {
            fprintf(stderr, "Error: expected 32 bit setwords\n");
            exit(1);
        }
    } else {
        fprintf(stderr, "64 bit mode\n");
        if(sizeof (setword) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit setwords -- sorry.\n");
            exit(1);
        }
        if(sizeof (unsigned long long int) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
            exit(1);
        }
    }
}

/******************************************************************************/

static void
print_help(char * argv0) {
    fprintf(stderr, "GenHypohamiltonian version %s\n", VERSION);
    fprintf(stderr, "Usage: %s <number_of_vertices> [options]\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Valid options are:\n");
    fprintf(stderr, "  g<x>: Searches for graphs with girth at least <x>.\n");
    fprintf(stderr, "  planar: Only output planar graphs.\n");
    fprintf(stderr, "  maxdeg <x>: Only output graphs with max degree at most 3.\n");
    fprintf(stderr, "  mod <rest> <modulo>: Splits the generation in <modulo> (more or less equally big) parts. Here part <rest> will be executed.\n");    
    fprintf(stderr, "  almostcomplete: Generate almost hypohamiltonian graphs (is exhaustive).\n");
    fprintf(stderr, "  almostspecial: Heuristic which generates almost hypohamiltonian graphs (is faster than almostcomplete, but may be incomplete).\n");
    fprintf(stderr, "  platypus: Generate platypus graphs (is exhaustive).\n");
}

/******************************************************************************/

/*
 * 
 */
int main(int argc, char** argv) {
    
    perform_wordsize_checks();
    nauty_check(WORDSIZE, MAXM, MAXN, NAUTYVERSIONID);

    if(argc == 2 && argv[1][0] == '-' && argv[1][1] == 'h') {
        print_help(argv[0]);
        exit(1);
    }

    if(argc < 2) {
        fprintf(stderr, "Error: invalid number of arguments. At least the number of vertices must be specified.\n");
        fprintf(stderr, "Usage: %s <number_of_vertices> [options]\n", argv[0]);
        fprintf(stderr, "Execute '%s -h' for more extensive help.\n", argv[0]);
        exit(1);
    } else {
        if(sscanf(argv[1], "%d", &maxnv)) {
            if(maxnv > MAXN) {
                if(MAXN == 32) {
                    fprintf(stderr, "Error: maximum number of vertices is 32 in this version.\n");
                    fprintf(stderr, "Compile with option -DWORDSIZE=64 in order to be able to generate graphs with a higher order.\n");
                    fprintf(stderr, "Or alternatively use the command 'make 64bit'\n");
                    exit(1);
                } else {
                    fprintf(stderr, "Error: number of vertices is too big (limit is %d).\n", MAXN);
                    exit(1);
                }
            }
            if(MAXN == 64 && maxnv <= 32) {
                fprintf(stderr, "Info: it is recommended to use the 32 bit version for generating graphs with order <= 32\n");
            }

        } else {
            fprintf(stderr, "Error: number of vertices should be an integer.\n");
            exit(EXIT_FAILURE);
        }

        int i;
        for(i = 2; i < argc; i++) {
            switch(argv[i][0]) {
                case 'm':
                {
                    if(strcmp(argv[i], "mod") == 0) {
                        if(i + 2 < argc) {
                            i++;
                            rest = atoi(argv[i]);
                            i++;
                            modulo = atoi(argv[i]);
                            if(rest >= modulo) {
                                fprintf(stderr, "Error: rest (%d) must be smaller than modulo (%d).\n", rest, modulo);
                                exit(1);
                            }
                        } else {
                            fprintf(stderr, "Error: option 'mod': missing rest or modulo\n");
                            exit(1);
                        }
                        fprintf(stderr, "Warning: splitting doesn't work very well with this program (a lot of isomorphic copies). Only works well with large mod.\n");
                    } else if(strcmp(argv[i], "maxdeg") == 0) {
                        impose_max_deg = 1;
                        if(i + 1 >= argc) {
                            fprintf(stderr, "Error: option 'maxdeg': missing value\n");
                            exit(1);
                        }
                        i++;
                        max_deg = atoi(argv[i]);
                        if(max_deg < 3) {
                            fprintf(stderr, "Error: maxdeg must be at least 3\n");
                            exit(1);
                        }
                        fprintf(stderr, "Info: only generating graphs with maxdeg at most %d\n", max_deg);
                    } else if(strcmp(argv[i], "minmaxdeg") == 0) {
                        impose_min_max_deg = 1;
                        if(i + 1 >= argc) {
                            fprintf(stderr, "Error: option 'maxdeg': missing value\n");
                            exit(1);
                        }
                        i++;
                        min_max_deg = atoi(argv[i]);
                        if(min_max_deg < 4) {
                            fprintf(stderr, "Error: minmaxdeg must be at least 4\n");
                            exit(1);
                        }
                        fprintf(stderr, "Info: only generating graphs with min maxdeg at most %d\n", min_max_deg);
                    }
                    else {
                        fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                        fprintf(stderr, "Execute '%s -h' for a list of possible options.\n", argv[0]);
                        exit(1);
                    }
                    break;
                }    
                case 'g':
                {
                    //Only accepts graphs which are NOT k-colourable (on last level)
                    girth = atoi(argv[i] + 1);
                    if(girth < 4) {
                        fprintf(stderr, "Error: when using girth option, it must be at least 4!\n");
                        exit(1);
                    }
                    test_girth = 1;
                    test_if_trianglefree = 1;
                    if(girth == 4)
                        test_girth = 0; //Do triangle-free instead
                    
                    fprintf(stderr, "Info: only outputting graphs with girth >= %d\n", girth);
                    
                    if(girth > 8) {
                        fprintf(stderr, "Error: only support for girth <= 8 at the moment!\n");
                        exit(1);
                    }
                    
                    break;
                }     
                case 'p':
                {
                    if(strcmp(argv[i], "planar") == 0) {
                        test_if_planar = 1;
                        fprintf(stderr, "Info: only generating graphs which are planar\n");
                    } else if(strcmp(argv[i], "platypus") == 0) {
                        test_platypus = 1;
                        fprintf(stderr, "Info: only generating platypus graphs\n");
                    }
                    break;
                }       
                case 'a':
                {
                    if(strcmp(argv[i], "almostspecial") == 0) {
                        test_almost_hypoham_special = 1;
                        fprintf(stderr, "Info: only generating graphs which are almost hypohamiltonian\n");
                        fprintf(stderr, "Warning: this is an incomplete hack which doesn't allow the special vertex to have larger degree than the hub!\n");
                    } else if(strcmp(argv[i], "almostcomplete") == 0) {
                        test_almost_hypoham_special = 1; //Must also be set!
                        test_almost_hypoham_complete = 1;
                        fprintf(stderr, "Info: only generating graphs which are almost hypoham\n");
                        fprintf(stderr, "Info: this allows one vertex to have a larger degree than the hub!\n");
                        fprintf(stderr, "Also it has not yet been tested if all obstructions are also valid for almost hypoham!\n");
                    } else {
                        fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                        fprintf(stderr, "Execute '%s -h' for a list of possible options.\n", argv[0]);
                        exit(1);                        
                    }
                    break;
                }                   
                default:
                {
                    fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                    fprintf(stderr, "Execute '%s -h' for a list of possible options.\n", argv[0]);
                    exit(1);
                }
            }
        }
    }
    
    if(test_platypus && (test_almost_hypoham_special || test_almost_hypoham_complete)) {
        fprintf(stderr, "Error: cannot use option platypus and almost hypoham at the same time!\n");
        exit(1);
    }
    
    //For the timing
    struct tms TMS;

    int i;
    for(i = 0; i < argc; i++)
        fprintf(stderr, "%s ", argv[i]);
    fprintf(stderr, "\n");
    
    //Reset for ham test (forced edges not used)
    for(i = 0; i <= maxnv; i++) {
        anzahl_forced[i] = 0;
        int j;
        for(j = 0; j < maxnv; j++) forced_edges[i][j] = 0;
    }        
    
    //Start construction
    initialize_splitting();
    
    init_construction();

    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;
    
    //if(test_if_trianglefree)
    //    fprintf(stderr, "Times graph is triangle-free: %llu (%.2f%%)\n", times_trianglefree, (double) times_trianglefree / (times_trianglefree + times_not_trianglefree) * 100);
    
    fprintf(stderr, "Times graph is non-hamiltonian: %llu (%.2f%%)\n", times_not_hamiltonian, (double) times_not_hamiltonian / (times_not_hamiltonian + times_hamiltonian) * 100);

    if(test_almost_hypoham_complete)    
        fprintf(stderr, "Times could prune because for except vertex v, G-v is ham: %llu (%.2f%%)\n", times_except_vertex_ham, (double) 100 * times_except_vertex_ham / (times_except_vertex_ham + times_except_vertex_not_ham));
    
    if(test_if_planar)
        fprintf(stderr, "Times graph is planar: %llu (%.2f%%)\n", times_planar, (double) times_planar / (times_planar + times_not_planar) * 100);    
    
    fprintf(stderr, "Times contains type A obstruction (given W): %llu (%.2f%%)\n", times_contains_type_A_obstruction, (double) times_contains_type_A_obstruction / (times_contains_type_A_obstruction + times_doesnt_contain_type_A_obstruction) * 100);
    
    fprintf(stderr, "Times contains deg 2 vertex: %llu (%.2f%%)\n", times_contains_deg_two_vertex, (double) times_contains_deg_two_vertex / (times_contains_deg_two_vertex + times_doesnt_contain_deg_two_vertex) * 100);
    
    fprintf(stderr, "Times contains type C obstruction: %llu (%.2f%%)\n", times_contains_type_C_obstruction, (double) times_contains_type_C_obstruction / (times_contains_type_C_obstruction + times_doesnt_contain_type_C_obstruction) * 100);        
    
    fprintf(stderr, "Times contains deg 3 triangle vertex: %llu (%.2f%%)\n", times_contains_deg_three_triangle_vertex, (double) times_contains_deg_three_triangle_vertex / (times_contains_deg_three_triangle_vertex + times_doesnt_contain_deg_three_triangle_vertex) * 100);

    fprintf(stderr, "Times contains type B obstruction: %llu (%.2f%%)\n", times_contains_type_B_obstruction, (double) times_contains_type_B_obstruction / (times_contains_type_B_obstruction + times_doesnt_contain_type_B_obstruction) * 100);    
    
    fprintf(stderr, "Times no lemmas applicable: %llu (%.2f%%)\n", times_no_lemmas_applicable, (double) times_no_lemmas_applicable / num_graphs_generated_noniso * 100);
    
    fprintf(stderr, "Max num edges added: %d\n", max_num_edges_added);

    fprintf(stderr, "Num non-iso graphs gen with %d vertices: %llu (incl. iso: %llu), noniso: %.2f%%\n", maxnv,
            num_graphs_generated_noniso, num_graphs_generated_noniso + num_graphs_generated_iso,
            (double) num_graphs_generated_noniso / (num_graphs_generated_noniso + num_graphs_generated_iso) * 100);

    if(test_platypus)
        fprintf(stderr, "Number of platypus graphs written: %llu\n", num_graphs_written);
    else {
        if(!test_almost_hypoham_special)
            fprintf(stderr, "Number of hypohamiltonian graphs written: %llu\n", num_graphs_written);
        else if(test_almost_hypoham_complete)
            fprintf(stderr, "Number of almost hypohamiltonian graphs written (complete) (could contain isomorphic copies!): %llu\n", num_graphs_written);
        else
            fprintf(stderr, "Number of almost hypohamiltonian graphs written (special): %llu\n", num_graphs_written);
    }
    fprintf(stderr, "Total CPU time: %.1f seconds.\n", (double) (savetime / time_factor));

    return (EXIT_SUCCESS);
    
}